@@ -9,4 +9,9 @@ from .serialize_wrapper import serialize_wrapper
 from .extractors import LabboxEphysRecordingExtractor, LabboxEphysSortingExtractor
 from .workspace_list import get_workspace_list, set_workspace_list, add_workspace_to_list
 from .extractors import H5SortingExtractorV1
-from .extractors.h5extractors.h5recordingextractorv1 import H5RecordingExtractorV1
\ No newline at end of file
+from .extractors.h5extractors.h5recordingextractorv1 import H5RecordingExtractorV1
+from .extractors.subrecording import subrecording
+from .extractors.subsorting import subsorting
+from .extractors.wrapperrecordingextractor import create_recording_from_old_extractor
+# from .SpikeAmplitudes import SpikeAmplitudes
+# from .SpikeExplorer import SpikeExplorer
\ No newline at end of file
