@@ -1,2 +1,2 @@
-from .prepare_snippets_h5 import prepare_snippets_h5
+from ._prepare_snippets_h5 import prepare_snippets_h5
 from .get_unit_waveforms_from_snippets_h5 import get_unit_waveforms_from_snippets_h5
\ No newline at end of file
